from pathlib import Path
import subprocess
import re

from bs4 import BeautifulSoup, NavigableString

from ._utils import get_main_texfile


FNAME_XML = "paper.xml"
FNAME_JATS = "paper.jats.xml"


def tex2xml(dpath_source: Path, title: str, verbose: bool) -> Path:
    dpath_work = dpath_source.parent
    fpath_xml = dpath_work / FNAME_XML
    fpath_jats = dpath_work / FNAME_JATS

    result = subprocess.run(
        ["which", "latexml"],
        capture_output=True,
        text=True,
    )
    if not result.stdout.strip():
        raise FileNotFoundError(
            "Could not find the `latexml` command. Please refer to "
            "this guide for installing LaTeXML: "
            "https://github.com/misya11p/arxiv2md"
        )

    fpath_tex = get_main_texfile(dpath_source, title)
    if not fpath_tex:
        raise FileNotFoundError(f"Could not find the main .tex file")

    command_latexml = [
        "latexml",
        fpath_tex,
        f"--dest={fpath_xml}"
    ]
    command_latexmlpost = [
        "latexmlpost",
        fpath_xml,
        "--format=jats",
        "--nographicimages",
        "--nodefaultresources",
        "--nopictureimages",
        "--nomathimages",
        f"--dest={fpath_jats}",
    ]
    subprocess_options = {
        "cwd": dpath_work,
        "stdout": None if verbose else subprocess.DEVNULL,
        "stderr": None if verbose else subprocess.DEVNULL,
    }
    subprocess.run(command_latexml, **subprocess_options)
    subprocess.run(command_latexmlpost, **subprocess_options)
    return fpath_jats


# ======================================================================
# The following was primarily written by Claude code, with minor
# modifications made by the developer


class JATSConverter:
    def __init__(self, dpath_source: Path):
        fpath_jats = dpath_source / FNAME_JATS
        with open(fpath_jats, "r", encoding="utf-8") as f:
            content = f.read()
        self.soup = BeautifulSoup(content, "xml")
        self._clear()

    def _clear(self):
        self.ref_counter = 0
        self.references = {}
        self.output = []

    def convert_to_md(self):
        self._clear()

        self._extract_title()
        self._extract_abstract()
        self._extract_body()
        self._extract_references()
        self._format_references()

        markdown_content = "\n".join(self.output)
        if markdown_content[-1] != "\n":
            markdown_content += "\n"

        return markdown_content

    @staticmethod
    def _clean_text(text):
        if not text:
            return ""
        text = text.strip()
        text = re.sub(r"\s+", " ", text)
        text = re.sub(r"\[\[(\^.*?)\]\]", r"[\1]", text)
        return text

    @staticmethod
    def _clean_bibid(bibid):
        return bibid.replace("bib.bib", "")

    @staticmethod
    def _clean_fnid(fn_id):
        return fn_id.replace("id", "fn")

    def _extract_title(self):
        title = self.soup.find("article-title")
        if title:
            self.output.append(f"# {self._clean_text(title.get_text())}\n")

    def _extract_abstract(self):
        abstract = self.soup.find("abstract")
        if abstract:
            self.output.append("## Abstract\n")
            for p in abstract.find_all("p"):
                self.output.append(self._process_paragraph(p))
            self.output.append("")

    def _extract_body(self):
        body = self.soup.find("body")
        if body:
            for section in body.find_all("sec", recursive=False):
                self._process_section(section, level=2)

    def _extract_references(self):
        ref_list = self.soup.find("ref-list")
        if not ref_list:
            return None

        for ref in ref_list.find_all("ref"):
            ref_id = ref.get("id", "")
            ref_id = self._clean_bibid(ref_id)
            citation = ref.find("mixed-citation") or ref.find("element-citation")
            if citation:
                ref_text = self._format_reference(citation)
                self.references[ref_id] = ref_text

    def _format_references(self):
        if not self.references:
            return

        self.output.append("## References\n")
        max_digit = len(max(self.references.keys(), key=len))
        for ref_id, ref_text in sorted(
            self.references.items(),
            key=lambda x: re.sub(r"\d+", lambda m: m.group(0).zfill(max_digit), x[0])
        ):
            self.output.append(f"[^{ref_id}]: {ref_text}")
        self.output.append("")

    def _process_section(self, section, level):
        title = section.find("title")
        if title:
            title_text = self._clean_text(title.get_text())
            self.output.append(f"{'#' * level} {title_text}\n")

        for child in section.children:
            if child.name == "sec":
                self._process_section(child, level + 1)
            elif child.name == "p":
                self.output.append(self._process_paragraph(child))
                self.output.append("")
            elif child.name == "fig":
                self._process_figure(child)
            elif child.name == "table-wrap":
                self._process_table(child)
            elif child.name in ["disp-formula", "disp-formula-group"]:
                self._process_formula(child)

    def _process_paragraph(self, p):
        result = []
        for child in p.children:
            if isinstance(child, NavigableString):
                result.append(str(child))
            elif child.name == "xref":
                ref_text = self._process_reference(child)
                result.append(ref_text)
            elif child.name == "italic":
                result.append(f"*{self._clean_text(child.get_text())}*")
            elif child.name == "bold":
                result.append(f"**{self._clean_text(child.get_text())}**")
            elif child.name == "inline-formula":
                math_text = self._extract_math_text(child)
                result.append(f"${math_text}$")
            elif child.name == "fn":
                result.append(self._process_footnote(child))
            else:
                result.append(self._process_mixed_content(child))

        processed_text = "".join(result)
        processed_text = re.sub(
            r"\[\[(\d+)\](?:,\s*\[(\d+)\])*\]",
            lambda m: self._fix_citation_group(m.group(0)),
            processed_text
        )
        return self._clean_text(processed_text)

    def _fix_citation_group(self, citation_text):
        numbers = re.findall(r"\d+", citation_text)
        return f"[{', '.join(numbers)}]"

    def _process_mixed_content(self, elem):
        if isinstance(elem, NavigableString):
            return str(elem)

        if elem.name == "xref":
            return self._process_reference(elem)
        elif elem.name == "italic":
            return f"*{self._clean_text(elem.get_text())}*"
        elif elem.name == "bold":
            return f"**{self._clean_text(elem.get_text())}**"
        elif elem.name == "inline-formula":
            math_text = self._extract_math_text(elem)
            return f"${math_text}$"
        else:
            return self._clean_text(elem.get_text())

    def _process_reference(self, xref):
        rid = xref.get("rid", "")
        if rid.startswith("bib.bib"): # Reference to bibliography
            rid = self._clean_bibid(rid)
            return f"[^{rid}]"
        else: # Reference to figure/table
            return xref.get_text()

    def _format_reference(self, citation):
        authors = []
        title = ""
        year = ""
        source = ""

        for person_group in citation.find_all("person-group"):
            for name in person_group.find_all("name"):
                surname = name.find("surname")
                given_names = name.find("given-names")
                if surname:
                    author_name = surname.get_text()
                    if given_names:
                        initials = "".join([n[0] + "." for n in given_names.get_text().split() if n])
                        author_name = f"{surname.get_text()}, {initials}"
                    authors.append(author_name)

        article_title = citation.find("article-title")
        if article_title:
            title = self._clean_text(article_title.get_text())

        year_elem = citation.find("year")
        if year_elem:
            year = year_elem.get_text().strip()

        source_elem = citation.find("source")
        if source_elem:
            source = self._clean_text(source_elem.get_text())

        if authors and title:
            author_str = ", ".join(authors[:3])
            if len(authors) > 3:
                author_str += " et al."

            ref_parts = []
            ref_parts.append(author_str)
            if year:
                ref_parts.append(f"({year})")
            ref_parts.append(f"*{title}*")
            if source:
                ref_parts.append(source)

            return " ".join(ref_parts)

        return self._clean_text(citation.get_text())

    def _process_figure(self, fig):
        caption = fig.find("caption")
        if caption:
            caption_text = self._process_paragraph(caption.find("p"))
            self.output.append(f"Figure: {caption_text}")
            self.output.append("")

    def _process_table(self, table_wrap):
        caption = table_wrap.find("caption")
        if caption:
            caption_text = self._process_paragraph(caption.find("p"))
            self.output.append(f"Table: {caption_text}")
            self.output.append("")

    def _process_formula(self, formula):
        math_elem = formula.find("math")
        if math_elem:
            math_text = self._extract_math_text(math_elem)
            self.output.append(f"$$\n{math_text}\n$$")
            self.output.append("")

    def _process_footnote(self, fn):
        fn_id = fn.get("id", "")
        fn_id = self._clean_fnid(fn_id)
        fn_text = fn.get_text(strip=True)
        if fn_id:
            self.references[fn_id] = fn_text
            return f"[^{fn_id}]"

    def _extract_math_text(self, formula_elem):
        if not formula_elem:
            return ""

        math_elem = formula_elem.find("math")
        if not math_elem:
            math_elem = formula_elem

        alttext = math_elem.get("alttext")
        if alttext:
            alttext = self._clean_math_alttext(alttext)
            return alttext

        return "[math]"

    def _clean_math_alttext(self, alttext):
        if not alttext:
            return ""

        alttext = alttext.replace("&#10;", "")
        alttext = alttext.replace("\n", "")
        alttext = alttext.replace("%", "")
        alttext = re.sub(r"\s+", " ", alttext)
        return alttext.strip()



from pathlib import Path

import typer

from ._utils import extract_arxiv_id
from ._api import arxiv2md_cli


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
app = typer.Typer(add_completion=False, context_settings=CONTEXT_SETTINGS)


@app.command()
def cli(
    url: str = typer.Argument(
        help=(
            "The URL of the arXiv paper or the arXiv ID."
        ),
    ),
    fpath_output: str = typer.Option(
        None,
        "--output", "-o",
        help=(
            "The path to the output Markdown file. If None, the file "
            "will be named as `arxiv_<arxiv_id>.md`."
        ),
    ),
    yes: bool = typer.Option(
        False,
        "--yes", "-y",
        help=(
            "If True, the command will not prompt for confirmation "
            "when overwriting existing files or creating directories."
        ),
    ),
    dpath_source: str = typer.Option(
        None,
        "--source-dir", "-d",
        help=(
            "The directory to store the source files "
            "(e.g., .tex, .xml). If None, a temporary directory will "
            "be used."
        ),
    ),
    no_frontmatter: bool = typer.Option(
        False,
        "--no-frontmatter",
        help=(
            "The output Markdown file will not include frontmatter "
            "metadata."
        ),
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose", "-v",
        help=(
            "The command will print detailed logs of the process."
        ),
    ),
):
    stdout = fpath_output == "-"
    arxiv_id = extract_arxiv_id(url)

    if dpath_source:
        dpath_source = Path(dpath_source).resolve()
        if dpath_source.exists():
            if (not yes) and (len(list(dpath_source.iterdir())) >= 1):
                if not typer.confirm(
                    f"The directory `{dpath_source}` already "
                    "exists and is not empty. If you continue, its "
                    "contents may be overwritten. Do you want to "
                    "continue?",
                    default=False,
                ):
                    raise typer.Exit()
        else:
            dpath_source.mkdir(parents=True)

    if not stdout:
        fpath_output = fpath_output or f"arxiv_{arxiv_id.replace('.', '-')}.md"
        fpath_output = Path(fpath_output).resolve()

        if not fpath_output.parent.exists():
            if yes or typer.confirm(
                f"The directory `{fpath_output.parent}` does not "
                "exist. Do you want to create it?",
                default=True,
            ):
                fpath_output.parent.mkdir(parents=True)

        if not fpath_output.suffix:
            fpath_output = fpath_output.with_suffix(".md")

        if fpath_output.exists() and (not yes):
            if not typer.confirm(
                f"{fpath_output} already exists. Do you want to "
                "overwrite it?"
            ):
                raise typer.Exit()

    content_md = arxiv2md_cli(
        arxiv_id, dpath_source, not no_frontmatter, verbose
    )

    if stdout:
        print(content_md)
    else:
        with open(fpath_output, "w", encoding="utf-8") as f:
            f.write(content_md)
        print(f"Markdown file saved to `{fpath_output}`")



import re
from pathlib import Path
import tarfile
import json
from typing import Dict
from difflib import SequenceMatcher

import arxiv


DNAME_SOURCE_ARXIV = "source_arxiv_{arxiv_id}"
FNAME_METADATA = "metadata.json"


def extract_arxiv_id(url: str) -> str:
    match = re.search(r"(\d{4}\.\d{4,5})", url)
    if match:
        arxiv_id = match.group(1)
        return arxiv_id
    else:
        raise ValueError(
            f"Invalid input URL: {url}. Could not extract arXiv ID."
        )


def get_source(url: str, dpath_source: Path) -> str:
    arxiv_id = extract_arxiv_id(url)
    paper = next(arxiv.Client().results(arxiv.Search(id_list=[arxiv_id])))
    fpath_source = paper.download_source(dpath_source)

    dname_source_arxiv = DNAME_SOURCE_ARXIV.format(
        arxiv_id=arxiv_id.replace('.', '-')
    )
    dpath_source_arxiv = dpath_source / dname_source_arxiv
    with tarfile.open(fpath_source, mode="r:gz") as tar:
        tar.extractall(dpath_source_arxiv)

    metadata = {
        "arxiv_id": arxiv_id,
        "title": paper.title,
        "published": paper.published.strftime("%Y-%m-%d"),
        "authors": [author.name for author in paper.authors],
    }
    with open(dpath_source / FNAME_METADATA, "w", encoding="utf-8") as f:
        json.dump(metadata, f, ensure_ascii=False, indent=2)

    return dpath_source_arxiv, metadata


def _main_tex_score(fpath_tex, title: str) -> float:
    score = 0
    with open(fpath_tex, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if line.startswith(r"\documentclass"):
                score += 20
            elif line.startswith(r"\begin{document}"):
                score += 5
            elif line.startswith(r"\title"):
                title_latex = re.sub(
                    r"[{}]", "", line[len(r"\title"):].strip()
                )
                ratio = SequenceMatcher(None, title, title_latex).ratio()
                score += 5 * ratio
            elif line.startswith(r"\maketitle"):
                score += 3
    return score

def get_main_texfile(dpath_source: Path, title: str) -> Path | None:
    tex_files = [
        (f, _main_tex_score(f, title))
        for f in dpath_source.glob("*.tex")
    ]
    tex_files = [(f, s) for f, s in tex_files if s > 0]

    if not tex_files:
        return None
    main_tex_file = max(tex_files, key=lambda x: x[1])[0]
    return main_tex_file


def concat_metadata(markdown: str, metadata: Dict) -> str:
    authors = "\n".join(
        [f"  - \"{author}\"" for author in metadata["authors"]]
    )
    frontmatter = "\n".join([
        "---",
        f"title: \"{metadata['title']}\"",
        f"arxiv_id: \"{metadata['arxiv_id']}\"",
        f"published: \"{metadata['published']}\"",
        f"authors:\n{authors}",
        "---",
        "",
    ])
    return frontmatter + markdown



from pathlib import Path
import tempfile
from typing import Tuple, Dict

from ._utils import extract_arxiv_id, get_source, concat_metadata
from ._convert import tex2xml, JATSConverter


def _core_arxiv2md_cli(
    arxiv_id: str,
    dpath_source: Path,
    verbose: bool,
) -> str:
    from halo import Halo

    with Halo(
        text=f"Get source for arXiv:{arxiv_id}",
        spinner="dots",
    ) as spinner:
        dpath_source_arxiv, metadata = get_source(arxiv_id, dpath_source)
        spinner.succeed()

    with Halo(
        text=f"Convert to Markdown",
        spinner="dots",
        enabled=not verbose,
    ) as spinner:
        if verbose:
            print("Converting to Markdown")
        tex2xml(dpath_source_arxiv, metadata["title"], verbose)
        converter = JATSConverter(dpath_source)
        content_md = converter.convert_to_md()
        spinner.succeed()

    return content_md, metadata


def _core_arxiv2md(
    arxiv_id: str,
    dpath_source: Path,
    verbose: bool
) -> Tuple[str, Dict]:
    dpath_source_arxiv, metadata = get_source(arxiv_id, dpath_source)
    tex2xml(dpath_source_arxiv, metadata["title"], verbose)
    converter = JATSConverter(dpath_source)
    content_md = converter.convert_to_md()
    return content_md, metadata


def arxiv2md_cli(
    url: str,
    dpath_source: str | Path | None,
    frontmatter: bool,
    verbose: bool,
) -> str:
    arxiv_id = extract_arxiv_id(url)

    if dpath_source:
        content_md, metadata = _core_arxiv2md_cli(
            arxiv_id, dpath_source, verbose
        )
    else:
        with tempfile.TemporaryDirectory() as tempdir:
            dpath_source = Path(tempdir)
            content_md, metadata = _core_arxiv2md_cli(
                arxiv_id, dpath_source, verbose
            )

    if frontmatter:
        content_md = concat_metadata(content_md, metadata)

    return content_md


def arxiv2md(
    url: str,
    dpath_source: str | Path | None = None,
    frontmatter: bool = False,
    verbose: bool = False,
) -> Tuple[str, Dict]:
    """
    Convert an arXiv paper to Markdown.

    Args:
        url (str): The URL of the arXiv paper or the arXiv ID.
        dpath_source (str | None, optional): The directory path to store
            the source files (e.g., .tex, .xml). If None, a temporary
            directory will be used. Defaults to None.
        frontmatter (bool, optional): If True, the output Markdown
            will include frontmatter metadata. Defaults to True.
        verbose (bool, optional): If True, print detailed logs during
            the conversion process. Defaults to False.

    Returns:
        Tuple[str, Dict]: A tuple containing the Markdown content and
            metadata. The metadata includes the arXiv ID, title,
            published date, and authors.
    """
    arxiv_id = extract_arxiv_id(url)

    if dpath_source:
        dpath_source = Path(dpath_source).resolve()
        if dpath_source.exists():
            if len(list(dpath_source.iterdir())) >= 1:
                raise FileExistsError(
                    f"The directory `{dpath_source}` already exists "
                    "and is not empty. Please specify an empty or "
                    "non-existing directory."
                )
        else:
            dpath_source.mkdir(parents=True, exist_ok=True)
        content_md, metadata = _core_arxiv2md(
            arxiv_id, dpath_source, verbose
        )
    else:
        with tempfile.TemporaryDirectory() as tempdir:
            dpath_source = Path(tempdir)
            content_md, metadata = _core_arxiv2md(
                arxiv_id, dpath_source, verbose
            )

    if frontmatter:
        content_md = concat_metadata(content_md, metadata)

    return content_md, metadata
